unit unit_robot;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, math;

type
  TPos = record
    x,y,th: Double;
  end;

  TVel = record
    v,w: Double;
  end;

  TStepPos = record
    mot: Byte;
    steps: Int16;
    th_ref,th: Double;
    acel: Integer;
  end;

  TStepVel = record
    mot: Byte;
    steps,lastSteps: UInt16;
    deltaSteps: Int16;
    w_ref,w_ctrl,w: Double;
    acel: Integer;
  end;

  TRobot = record
    D,L,ThOff: double;
    MaxTh,MinTh: double;
    pos: TPos;
    vel: TVel;
    M0: TStepPos;
    M1: TStepVel;
  end;

const
  NUM_MOTS = 2;
  // Steppers proprieties
  STEPS_PER_180DEG: array [0..1] of Integer = ( 1600 , 800 ); // #STEPS / 180º
  K_TIME = 2.6048411231406;                                   // Control period of the steppers speed

var
  robot: TRobot;
         
function NormalizeAngle(ang: double): double;
function PosStepsToPosAngular(posSteps: Integer; mot: Byte): Double;
function PosAngularToPosSteps(PosAngular: Double; mot: Byte): Integer;
function SpeedStepsToSpeedAngular(speedSteps: Integer; mot: Byte): Double;
function SpeedAngularToSpeedSteps(speedAngular: Double; mot: Byte): Integer;

procedure RobotInit;
procedure UpdateOdometry(deltaTime: double);
procedure InverseKinematics(vnom,wnom: Double);
procedure ResetOdometry;
procedure StopRobot;

implementation

function NormalizeAngle(ang: double): double;
var
  ang_normalized: double;
begin
  ang_normalized := ang;
  while ang_normalized > pi do begin
    ang_normalized := ang_normalized - 2*pi;
  end;
  while ang_normalized <= -pi do begin
    ang_normalized := ang_normalized + 2*pi;
  end;
  NormalizeAngle := ang_normalized;
end;

function PosStepsToPosAngular(posSteps: Integer; mot: Byte): Double;
begin
  Result := NormalizeAngle(pi*posSteps/STEPS_PER_180DEG[mot]);
end;

function PosAngularToPosSteps(PosAngular: Double; mot: Byte): Integer;
begin
  Result := Round(PosAngular*STEPS_PER_180DEG[mot]/pi);
end;

function SpeedStepsToSpeedAngular(speedSteps: Integer; mot: Byte): Double;
begin
  Result := speedSteps*pi/(K_TIME*STEPS_PER_180DEG[mot]);
end;

function SpeedAngularToSpeedSteps(speedAngular: Double; mot: Byte): Integer;
begin
  Result := Round(speedAngular*K_TIME*STEPS_PER_180DEG[mot]/pi);
end;




procedure RobotInit;
const
  MAX_WH_TH =  90;
  MIN_WH_TH = -90;
begin
  robot.M0.mot  := 0;  
  robot.M0.acel := 100;
  robot.M1.mot  := 1;
  robot.M1.acel := 100;
  robot.D := 0.065;
  robot.L := 0.18;
  robot.ThOff := 0;
  robot.MaxTh := DegToRad(MAX_WH_TH);  
  robot.MinTh := DegToRad(MIN_WH_TH);
end;

procedure UpdateOdometry(deltaTime: double);
var deltaWh,deltaD,deltaTh: double;
begin
  // Wheel displacement: steps count
  robot.M1.deltaSteps := Int16(robot.M1.steps - robot.M1.lastSteps);
  robot.M1.lastSteps  := UInt16(robot.M1.steps);
  robot.M1.w := pi*robot.M1.deltaSteps/(deltaTime*STEPS_PER_180DEG[1]);
  // Robot displacement
  deltaWh := pi*robot.D*robot.M1.deltaSteps/(2*STEPS_PER_180DEG[1]);
  deltaD  := deltaWh * cos(robot.M0.th + robot.ThOff);
  deltaTh := deltaWh * sin(robot.M0.th + robot.ThOff) / robot.L;
  // Odometry pose
  robot.pos.x  := robot.pos.x  + deltaD*cos(robot.pos.th + deltaTh/2);
  robot.pos.y  := robot.pos.y  + deltaD*sin(robot.pos.th + deltaTh/2);
  robot.pos.th := robot.pos.th + deltaTh;
  // Robot velocity
  robot.vel.v := deltaD  / deltaTime;
  robot.vel.w := deltaTh / deltaTime;
end;

procedure InverseKinematics(vnom,wnom: Double);
const MAX_TH_ERROR = 2;
      MAX_M1_ACEL  = 0.01;
begin
  // Inverse kinematics
  // - steering angle:
  if ((Abs(wnom) < 1e-4) AND (Abs(vnom) < 1e-4)) then begin  // Robot stopped
    robot.M0.th_ref := robot.M0.th_ref;            // ... maintains the previous steering angle
  end else if ((Abs(vnom) < 1e-4) AND (wnom > 0)) then begin
    robot.M0.th_ref :=  pi/2;
  end else if ((Abs(vnom) < 1e-4) AND (wnom < 0)) then begin
    robot.M0.th_ref := -pi/2;
  end else begin
    robot.M0.th_ref := NormalizeAngle(ArcTan(robot.L * wnom / vnom) - robot.ThOff);
  end;
  // - steering angle restrictions
  if (robot.M0.th_ref > robot.MaxTh) then
    robot.M0.th_ref := robot.MaxTh;
  if (robot.M0.th_ref < robot.MinTh) then
    robot.M0.th_ref := robot.MinTh;
  // - driving motor
  if ((RadToDeg(Abs(Abs(robot.M0.th_ref) - pi/2)) < MAX_TH_ERROR) AND (RadToDeg(Abs(Abs(robot.M0.th) - pi/2)) < MAX_TH_ERROR)) then begin
    robot.M1.w_ref := Abs(wnom)/(robot.D*robot.L);
  end else begin
    robot.M1.w_ref := 2*vnom/(robot.D*cos(robot.M0.th_ref + robot.ThOff));
  end;
  // - steering angle restrictions
  if (RadToDeg(NormalizeAngle(abs(robot.M0.th - robot.M0.th_ref))) > MAX_TH_ERROR) then begin // (3rd for the robot not being stopped)
    vnom := 0;
    wnom := 0;
    StopRobot;
  end;

  // Acceleration control (driving wheel)
  //if (robot.M1.w < robot.M1.w_ref) then begin
  if (robot.M1.w_ctrl < robot.M1.w_ref) then begin
    //robot.M1.w_ctrl := robot.M1.w + MAX_M1_ACEL;
    robot.M1.w_ctrl += MAX_M1_ACEL;
    if (robot.M1.w_ctrl >= robot.M1.w_ref) then
      robot.M1.w_ctrl := robot.M1.w_ref;
  //end else if (robot.M1.w > robot.M1.w_ref) then begin
  end else if (robot.M1.w_ctrl > robot.M1.w_ref) then begin
    //robot.M1.w_ctrl := robot.M1.w - MAX_M1_ACEL;
    robot.M1.w_ctrl -= MAX_M1_ACEL;
    if (robot.M1.w_ctrl <= robot.M1.w_ref) then begin
      robot.M1.w_ctrl := robot.M1.w_ref;
    end;
  end;
end;

procedure ResetOdometry;
begin
  robot.pos.x  := 0;
  robot.pos.y  := 0;
  robot.pos.th := 0;
  robot.M1.lastSteps := 0;
  robot.M1.steps     := 0;
  robot.M1.deltaSteps:= 0;
end;

procedure StopRobot;
begin
  robot.M1.w_ctrl := 0;
  robot.M1.w_ref  := 0;
end;

end.

